# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# wdic-menu.pl
# $Id: wdic-menu.pl 1.3.3.5 2003/08/02 03:19:31 satomii Exp $
#
# This file is designed to be used from the wdic-fpw.pl script.
# DO NOT CHANGE THE ENCODING OF THIS FILE.
#
# Copyright (C) 2000-2003, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use strict;
use File::DosGlob 'glob';
use IO::File;
use Jcode;

require 'wdic-reform.pl';

my %LABEL_SJIS = (
	ref => 'b',
	index => 'O[vʎ^ꗗ',
	dir => 'ޕʎ^ꗗ',
	graph => '摜ꗗ',
	plugin => 'vOCꗗ',
);
my %LABEL_EUC = (
	bull_dir => Jcode::convert('', 'euc', 'sjis'),
	bull_word => Jcode::convert('', 'euc', 'sjis'),
	none => Jcode::convert('(YPȂ)', 'euc', 'sjis'),
);
my @copyright;

#
# void make_menu(void)
# writes the description about each groups in the current text context.
#
sub make_menu()
{
	my ($readme, $man, $question, $file, $title, %refs, @tags);
	local $_;

	&log("Writing menu...\n");

	#
	# readme.txt, required. the file existence check should have been done
	# in wdic-fpw.pl, but we use glob again to ignore case (if possible
	# and applicable).
	#
	$file = glob("./readme.txt") || "./readme.txt";
	$readme = &new_file_handle($file, 'r');
	while ($_ = $readme->getline()) {
		if (s/^([^\r\n]+[\r\n]+)$/$1/) {
			&write_link($fpwmenu, $_, 'README');
			last;
		}
	}

	#
	# WDICALL.MAN, required. note that the name of this file was WDIC.MAN
	# in older versions of WDIC.
	#
	$file = glob("$wdic_dir/WDICALL.MAN") || "$wdic_dir/WDICALL.MAN";
	$man = &new_file_handle($file, 'r');
	while ($_ = $man->getline()) {
		if (s/^\s*(ʐMp(?: *\(\) *)?̊bm)\s*([0-9]+[\r\n]+)$/$1 $2/) {
			$title = &reform_sjis($1, 0);
			$_ = &reform_sjis($_, 0);
			last;
		}
	}
	$title = 'ʐMp () ̊bm' if (!$title);
	&write_link($fpwmenu, "$_\n", 'MAN');

	#
	# references from *.MAN, optional.
	#
	foreach (keys(%{ $wdic->{group_order} })) {
		$file = glob("$wdic_dir/$_.MAN");
		if ($file || -f "$wdic_dir/$_.MAN") {
			$refs{$_} = $file ? $file : "$wdic_dir/$_.MAN";
		}
	}
	&write_link($fpwmenu, "$LABEL_SJIS{ref}\n", 'REF') if (%refs);

	#
	# QUESTION.TXT, optional. the recent version of WDIC does not contain
	# this file.
	#
	$file = glob("$wdic_dir/QUESTION.TXT") || "$wdic_dir/QUESTION.TXT";
	eval { $question = &new_file_handle($file, 'r') };
	if ($question) {
		while ($_ = $question->getline()) {
			if (s/^\s+([^\r\n]+[\r\n]+)$/$1/) {
				&write_link($fpwmenu, $_, 'QUESTION');
				last;
			}
		}
	}

	#
	# lists of entries in this book. make_dir_list must be called later to
	# generate the target page.
	#
	@tags = glob("$wdic_dir/*.TAG");
	if (@tags) {
		&write_link($fpwmenu, "$LABEL_SJIS{index}\n", 'GROUP');
	}
	if (keys(%{ $wdic->{wdic_dirs} })) {
		&write_link($fpwmenu, "$LABEL_SJIS{dir}\n", 'DIR');
	}

	#
	# a list of plugins, if --cgraphs and/or --plugin is set.
	# make_plugin_list must be called later to generate the target pages.
	#
	if ($opt_cgraphs) {
		&write_link($fpwmenu, $LABEL_SJIS{graph} . "\n", 'GRAPH');
	}
	if ($opt_plugin) {
		&write_link($fpwmenu, $LABEL_SJIS{plugin} . "\n", 'PLUGIN');
	}

	#
	# create target entries.
	#
	&write_readme($readme);
	$readme->close();
	&write_man($man, $title);
	$man->close();
	&write_book_references(%refs) if (%refs);

	if ($question) {
		&write_question($question);
		$question->close();
	}

	&make_group_index(@tags) if (@tags);
	&make_dir_index() if ($wdic->{wdic_dirs});
}

#
# void new_menu_context(char *title, char *tag)
# creates a new menu context with the given title and tag. the indentation
# of the context is set to 2 by default.
#
sub new_menu_context($$)
{
	my ($title, $tag) = @_;

	$fpwmenu->new_context()
		|| die &fpwerr("Unable to add menu context", $fpwmenu);
	if ($fpwmenu->can('add_entry_tag')) {
		$fpwmenu->add_entry_tag($tag)
			|| die &fpwerr("Unable to add entry tag: $tag", $fpwmenu);
	} else {
		$fpwmenu->add_tag($tag)
			|| die &fpwerr("Unable to add entry tag: $tag", $fpwmenu);
	}
	return if (!$title);

	if (!$fpwmenu->add_keyword_start()
			|| !$fpwmenu->add_text(Jcode::convert($title, 'euc', 'sjis'))
			|| !$fpwmenu->add_keyword_end()
			|| !$fpwmenu->add_newline()) {
		die &fpwerr("Unable to write menu title", $fpwmenu);
	}
}

##------------------------------------------------------------------------
## MENU WRITER FUNCTIONS

sub write_readme($)
{
	my $handle = $_[0];
	local $_;

	&log("readme.txt\n");
	$handle->seek(0, IO::File::SEEK_SET);
	while ($_ = $handle->getline()) {
		last if (s/^([^\r\n]+)[\r\n]+$/$1/);
	}
	&new_menu_context($_, 'README');
	while ($_ = $handle->getline()) {
		&write_sjis($fpwmenu, $_);
	}
}

#
# void write_man(handle, char *title)
# create a new menu entry from the WDICALL.MAN file.
#
sub write_man($$)
{
	my ($handle, $title) = @_;
	my ($index, $next, $prev);
	my $format = 'MAN%02x';
	local $_;

	&log("$wdic_dir/WDICALL.MAN\n");
	&new_menu_context('', 'MAN');

	$index = 0;
	while ($_ = $handle->getline()) {
		if (s/^\-+  (.+)  \-+[\r\n]+$/$1/) {
			$index++;
			$_ = &reform_sjis($_, 1) if (/\x81[\x69\x6a]/);
			&write_link($fpwmenu, "$_\n", sprintf($format, $index));
		}
	}

	$index = 0;
	$handle->seek(0, IO::File::SEEK_SET);
	while ($_ = $handle->getline()) {
		last if (s/^\-+  (.+)  \-+[\r\n]+$/$1/);
	}

	do {
		print "  $_\n";
		$index++;
		&new_menu_context($_, sprintf($format, $index));
		$fpwmenu->add_indent_level(2);

		# dispatch the section to the writer-function.
		if ('e̓F' eq $_) {
			$next = &write_man_section_features($handle);
		} elsif ('tH[}bg̑' eq $_) {
			$next = &write_man_section_format($handle);
		} elsif ('iɂ' eq $_) {
			$next = &write_man_section_pos($handle);
		} elsif ('ǂ݂ɂ' eq $_ || 'ɂ' eq $_) {
			$next = &write_man_section_reading($handle);
		} elsif ('̑̋K' eq $_) {
			$next = &write_man_section_rules($handle);
		# 1.3c: removed since 2003-2 refers to other manual files.
		#} elsif ('b' eq $_) {
		#	$next = &write_man_section_references($handle);
		} elsif ('ₗ' eq $_ || '' eq $_) {
			$next = &write_man_section_versions($handle);
		} elsif ('M' eq $_) {
			$next = &write_man_section_environment($handle);
		} elsif ('t' eq $_ || 'A' eq $_) {
			$next = &write_man_section_credit($handle);
		} else {
			# 1.3: the rest is not (yet) separately handled.
			$next = &write_man_section($_, $handle);
		}
		$next = &reform_sjis($next, 1) if ($next =~ /\x81[\x69\x6a]/);

		# menu footer (navigation).
		$fpwmenu->add_indent_level(2);
		if ('' ne $prev) {
			&write_sjis($fpwmenu, '');
			&write_link($fpwmenu, $prev, sprintf($format, $index - 1));
		}
		if ('' ne $next) {
			&write_sjis($fpwmenu, ($prev ? ', ' : '') . '');
			&write_link($fpwmenu, $next, sprintf($format, $index + 1));
		}
		&write_sjis($fpwmenu, ', ');
		&write_link($fpwmenu, "$title\n", 'MAN');

		$prev = $_;
		$_ = $next;
	} while ($_);
}

#
# writes the "kaku-jisho no tokushoku" section into the current menu
# context, then returns the first line of the next section.
#
sub write_man_section_features($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	my ($file, $cat);
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ +(\[[A-Z]+ *$SJIS_ANY+\])/$1/) {
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);

		} elsif (s/^ +([A-Z]+\.DIC) +([^ ]+) +//) {
			$file = $1 . (" " x (15 - length($1)));
			$cat = $2 . (" " x (20 - length($2)));
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, "$file$cat" . &reform_sjis($_, 1));
			$fpwmenu->add_nowrap_end();
			$newline = &write_newline($fpwmenu, 1);

		} elsif (s/^ +([A-Z]+\.[A-Z]+) +([^ ])/$2/) {
			$file = $1 . (" " x (15 - length($1)));
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, $file . &reform_sjis($_, 1));
			$fpwmenu->add_nowrap_end();
			$newline = &write_newline($fpwmenu, 1);

		} elsif (3 == $indent) {
			s/^ +//;
			$fpwmenu->add_indent_level(2);
			$indent = 2;
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			$prev = &reform_sjis($_, 1);
			&write_sjis($fpwmenu, $prev);
			if (length($prev) < 15 && $prev =~ /^$SJIS_ANY+$/) {
				$newline = &write_newline($fpwmenu, 2);
			} else {
				$newline = 0;
				next;
			}

		} else {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;
		}

		$prev = '';
	}

	$_;
}

#
# writes the "format sono-ta" section into the current menu context, then
# returns the first line of the next section.
#
sub write_man_section_format($)
{
	my ($handle) = @_;
	my ($newline, $prev) = (1, '');
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ {3}\x81[\xa1\x45] *//) {
			&write_sjis($fpwmenu, " $_");
			$newline = &write_newline($fpwmenu, 2);

		} elsif (s/^ {6} *(#|y$SJIS_ANY+z)//) {
			my $prefix = $1;
			if ('#' ne $prefix) {
				$prefix = &reform_sjis($prefix, 0);
				$_ = &reform_sjis($_, 0);
				s/^(>+) +/$1/;
			}
			$fpwmenu->add_indent_level(3);
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, "$prefix$_");
			$fpwmenu->add_nowrap_end();
			&write_newline($fpwmenu, 1);
			while ($_ = $handle->getline()) {
				last if (/^[ \r\n]*$/);
				s/[ \r\n]+$//;
				s/^ +//;
				if ('#' eq $prefix) {
					$_ = (" " x 8) . &reform_sjis($_, 1);
				} else {
					$_ = &reform_sjis($_, 0);
					s/^(>+) +/$1/;
					$_ = (" " x length($prefix)) . $_;
				}
				$fpwmenu->add_nowrap_start();
				&write_sjis($fpwmenu, $_);
				$fpwmenu->add_nowrap_end();
				&write_newline($fpwmenu, 1);
			}
			$fpwmenu->add_indent_level(2);
			&write_newline($fpwmenu, 1);
			$newline = 2;

		} else {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;
		}

		$prev = '';
	}

	$_;
}

#
# writes the "hinshi-mei ni tsuite" section into the current menu context,
# then returns the first line of the next section.
#
sub write_man_section_pos($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			$newline = &write_newline($fpwmenu, 1) if (!$newline);
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev = '';
			next;

		} elsif ($prev) {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;

		} elsif (s/^ + *//) {
			&write_sjis($fpwmenu, " $_");
			$newline = &write_newline($fpwmenu, 2);
			$prev = '';
			next;

		} elsif (s/^ +()/$1/) {
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);
			$prev = '';
			next;

		} elsif (s/^  (\x84[\xab\xac\xaf])/$1/) {
			# table, nop here.

		} elsif (s/^ {6}((?:[ \|]+)?\+\@[^ ]+ *)//) {
			# semantic part-of-speech tree.
			my $name = $1;
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			if ($_) {
				$_ = &reform_sjis($_, 1);
				s/(\@$SJIS_ANY+) (ETC)/$1$2/g;
			}
			$_ = "$name$_";

		} elsif (s/^ {6}([ \|]+)$/$1/) {
			# semantic part-of-speech node, branch only. nop.

		} elsif (s/^ {6}( +\+[^ ]+)//) {
			# child node of extended part-of-speech.
			my $name = $1;
			$_ = &reform_sjis($_, 1) if ($_);
			$_ = $name . (' ' x (20 - length($name))) . $_;

		} elsif (s/^ {5} ?($SJIS_ANY+)$/$1/ && length($_) <= 10) {
			# parent node of extended part-of-speech.
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}

		} elsif (s/^ {8} *($SJIS_ANY+  +[\@\+])/$1/) {
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}

		} else {
			s/^ +//;
			$prev = &reform_sjis($_, 1);
			$prev =~ s/ +$//;
			&write_sjis($fpwmenu, $prev);
			$newline = 0;
			next;
		}

		$fpwmenu->add_nowrap_start();
		&write_sjis($fpwmenu, $_);
		$fpwmenu->add_nowrap_end();
		$newline = &write_newline($fpwmenu, 1);
		$prev = '';
	}

	$_;
}

#
# writes the "yomi ni tsuite" or "hatsuon ni tsuite" section into the
# current menu context, then returns the first line of the next section.
#
sub write_man_section_reading($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			$newline = &write_newline($fpwmenu, 1) if (!$newline);
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev = '';
			next;

		} elsif ($prev) {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/) {
				$fpwmenu->add_text(' ');
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;

		} elsif (s/^ {3}E/ / || s/^ {4}E($SJIS_ANY+)$/ $1/) {
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);
			$prev = '';
			next;

		} elsif (s/^ {10} *(E$SJIS_ANY+)$/$1/) {
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$_ = &reform_sjis($_, 1);

		} elsif (s/^ {8} *(\[\])/$1/
				|| s/^ {8} ?([^ ]+  )/$1/
				|| s/^ {6}( *[^ ]+  )/$1/
				|| s/^ {5}([^ ].*  )/$1/) {
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}

		} else {
			s/^ +//;
			s/^E//;
			$prev = &reform_sjis($_, 1);
			$prev =~ s/ +$//;
			&write_sjis($fpwmenu, $prev);
			$newline = 0;
			next;
		}

		$fpwmenu->add_nowrap_start();
		&write_sjis($fpwmenu, $_);
		$fpwmenu->add_nowrap_end();
		$newline = &write_newline($fpwmenu, 1);
		$prev = '';
	}

	$_;
}

#
# writes the "sono-ta no kitei" section into the current menu context, then
# returns the first line of the next section.
#
sub write_man_section_rules($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ E// || s/^  +E/ /) {
			# the first line or heading.
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);

		} else {
			s/^ +//;
			if (/  / || (!/,/ && !/\.$/)) {
				if (s/^((?:\x82[\x4f-\x58])+) +//) {
					my $mon = &reform_sjis($1, 1);
					$_ = $mon . (' ' x (13 - (length($mon)))) . $_;
				} elsif (s/^(\x81.) +//) {
					$_ = $1 . (' ' x 11) . &reform_sjis($_, 1);
				}
				if (3 != $indent) {
					$fpwmenu->add_indent_level(3);
					$indent = 3;
				}
				$fpwmenu->add_nowrap_start();
				&write_sjis($fpwmenu, $_);
				$fpwmenu->add_nowrap_end();
				$newline = &write_newline($fpwmenu, 1);
			} else {
				$_ = &reform_sjis($_, 1);
				s/ +$//;
				if ($prev) {
					if ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/) {
						$fpwmenu->add_text(" ");
					}
				} elsif (2 != $indent) {
					$fpwmenu->add_indent_level(2);
					$indent = 2;
				}
				&write_sjis($fpwmenu, $_);
				$prev = $_;
				$newline = 0;
				next;
			}
		}
		$prev = '';
	}

	$_;
}

#
# writes the "kiso bunken" section into the current menu context, then
# returns the first line of the next section.
#
sub write_man_section_references($)
{
	my ($handle) = @_;
	my ($newline, $prev) = (1, '');
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev = '';

		} elsif (s/^[\- ]+ +([^ ].*[^ ]) +[\- ]+$/ $1/
				|| s/^\[(($SJIS_ANY+|[^\[\]]+))\]/ $1/) {
			$_ = &reform_sjis($_, 1);
			&write_sjis($fpwmenu, $_);
			$newline = &write_newline($fpwmenu, 2);
			$prev = '';

		} elsif (s/^  (RFC [0-9]+ +)//) {
			$prev = "$1: " . &reform_sjis($_, 1);
			&write_newline($fpwmenu, 1) if (!$newline);
			&write_sjis($fpwmenu, $prev);
			$newline = 0;

		} elsif (s/^ *(E)/$1/) {
			s/((?: |$SJIS_ANY)\x81\x5e)([^ ])/$1 $2/g;
			s/  +/ /g;
			&write_newline($fpwmenu, 1) if (!$newline);
			$prev = &reform_sjis($_, 1);
			$prev =~ s/ +$//;
			&write_sjis($fpwmenu, $prev);
			$newline = 0;

		} elsif (s/^ *((.+))$/$1/ || s/^ *\<(.+)\>$/$1/) {
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);
			$prev = '';

		} elsif (s/^k(.+)l$/$1/) {
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);
			$prev = '';

		} elsif (s/^ {10} *//) {
			s/( \x81\x5e)([^ ])/$1 $2/g;	# ^
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			$fpwmenu->add_text(' ') if ($prev);
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;

		} elsif (!$prev && (4 <= length($_) || /^$SJIS_ANY+n$/)) {
			s/^ +//;
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);

		} else {
			s/( \x81\x5e)([^ ])/$1 $2/g;	# ^
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
				$fpwmenu->add_text(' ');
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
		}
	}

	$_;
}

#
# writes the "kaitei rireki" section into the current menu context, then
# returns the first line of the next section.
#
sub write_man_section_versions($)
{
	my ($handle) = @_;
	my ($next, $space, $max, @table, @comment);
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//) {
			$next = $_;
			last;
		} elsif (s/^ *//) {
			push(@comment, $_);
		} elsif (/^( +)-+[ \\-]+$/) {
			$space = length($1) if (!$space);
			push(@table, $_);
		} elsif (0 < scalar(@comment)) {
			s/^ +//;
			if (/^ *$/) {
				push(@comment, $_) if ($comment[$#comment] !~ /^$/);
			} else {
				$comment[$#comment] .= $_;
			}
		} elsif (/^ *$/) {
			push(@table, '') if (!@table || $table[$#table] !~ /^$/);
		} else {
			if (s/( [0-9,]+ +)([^0-9, ].+)$/$1/) {
				my $str = &reform_sjis($2, 1);
				$_ .= $str;
				$max = length($str) if ($max < length($str));
			}
			push(@table, $_);
		}
	}

	foreach (@table) {
		s/^ {$space}//;
		if ($_) {
			$fpwmenu->add_nowrap_start();
			if (s/^([\-= ]+ )([\-=])\2+$/$1/) {
				$fpwmenu->add_text($_ . ($2 x $max))
					|| die &death_cry("Unable to write text", $fpwmenu);
			} else {
				&write_sjis($fpwmenu, $_);
			}
			$fpwmenu->add_nowrap_end();
		}
		&write_newline($fpwmenu, 1);
	}

	foreach (@comment) {
		&write_sjis($fpwmenu, &reform_sjis($_, 1)) if ($_);
		&write_newline($fpwmenu, 1);
	}

	$next;
}

#
# writes the "shippitsu kankyou" section into the current menu context,
# then returns the first line of the next section.
#
sub write_man_section_environment($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	my $name;
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ {5}(?:  )?((?:[^ ]+ ?)+) {4} *([^ ])/$2/
				 || s/^ {20} *(ver)/$1/i) {
			if ($2) {
				$name = &reform_sjis($1, 1);
				$name .= (" " x (24 - length($name)));
			} else {
				$name = " " x 24;
			}
			if (s/(  )((?:[^ ]+ ?)+)$/$1/) {
				$_ = $_ . &reform_sjis($2, 1);
			}
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, "$name$_");
			$fpwmenu->add_nowrap_end();
			$newline = &write_newline($fpwmenu, 1);

		} elsif (s/^ {5}([^ ])/ $1/) {
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);

		} else {
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			s/^ +//;
			if ('fXNgbv' eq $_ || 'm[g' eq $_) {
				$prev = " $_";
			} else {
				$_ = &reform_sjis($_, 1);
				s/ +$//;
				if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
					$fpwmenu->add_text(" ");
				}
				$prev = $_;
			}
			&write_sjis($fpwmenu, $prev);
			$newline = 0;
			next;
		}

		$prev = '';
	}

	$_;
}

#
# char *write_man_section_credit($handle)
# converts the "okuduke" or "renraku saki" section into a menu entry.
#
sub write_man_section_credit($)
{
	my ($handle) = @_;
	my $newline = 1;
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);
		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
		} else {
			if (s/^ +([^ ]+) +([^ ])/$2/) {
				$_ = $1 . ' ' x (10 - length($1)) . &reform_sjis($_, 1);
				push(@copyright, $_);
			} else {
				s/^ +//;
			}
			&write_sjis($fpwmenu, "$_\n");
			$newline = 1;
		}
	}
	$_;
}

#
# char *write_man_section(char *title, $handle)
# converts a WDICALL.MAN section into a menu entry.
#
sub write_man_section($$)
{
	my ($title, $handle) = @_;
	my ($newline, $indent, $prev_indent, $prev_str) = (1, 2, 2, '');
	local $_;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-+$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev_str = '';
			next;

		} elsif ('T|[g' eq $title) {
			if (s/^ *(\(\x82[\x4f-\x58]\)) */$1 /) {
				$_ = &reform_sjis($_, 1);
				$indent = 2;
			} else {
				s/^ +//;
				s/([^\r\n])[\r\n]+$/$1/ if (!/^[\x20-\x7e]+$/);
				$indent = 3;
			}

		} elsif (s/^ *E// || s/^ *(\(\%[0-9]\) )/$1/) {
			# a line with bullet or notation index.
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev_str = '';
			s/([^\r\n]*)[\r\n]+$/$1/;
			$indent = 2;

		} elsif (s/^ *([0-9]\. *)/$1 /) {
			# also with an outline number.
			$newline = &write_newline($fpwmenu, 1) if (!$newline);
			$prev_str = '';
			s/([^\r\n]*)[\r\n]+$/$1/;

		} elsif (s/^ *([a-z]+:\/\/[\x21-\x7e]+[\r\n]+)$/  $1/) {
			# URL-only line.
			$indent = 3;

		} else {
			# default processing.
			s/^ +//;
			s/([^\r\n])[\r\n]+$/$1/;
			$indent = 2;
		}

		# write the current string to the menu context.
		if ($indent != $prev_indent) {
			$fpwmenu->add_indent_level($indent)
				|| die &death_cry("Unable to set indent level ($indent)",
					$fpwmenu);
			$prev_indent = $indent;
		}
		$newline = /[\r\n]+$/;
		if ($newline) {
			$prev_str = '';
		} else {
			$_ = &reform_sjis($_, 1);
			if ('' ne $prev_str && $prev_str !~ / $/ && !/^ /) {
				if ($prev_str !~ /$SJIS_ANY$/ || /^[\x21-\x7e]/) {
					$_ = " $_";
				}
			}
			$prev_str = $_;
		}
		&write_sjis($fpwmenu, $_);

		# more fix by my preference.
		if (/^/) {
			while ($newline < 2) {
				$fpwmenu->add_newline()
					|| die &death_cry("Failed to add newline", $fpwmenu);
				$newline++;
			}
		}
	}

	$_;
}

#
# void write_book_references((hash) files)
# creates bibliography entries.
#
sub write_book_references(@)
{
	my %refs = @_;
	my @keys = $wdic->sort_group_keys(keys(%refs));
	my (%entry, $key);
	local $_;

	&log("$LABEL_SJIS{ref}\n");
	&new_menu_context('', 'REF');

	foreach $key (@keys) {
		%entry = {};
		$entry{file} = $refs{$key};
		$entry{handle} = new IO::File($entry{file}, 'r')
			|| die "Unable to open file: $entry{file}";
		$entry{title} = $wdic->get_group_title($key);
		$entry{tag} = "REF_$key";

		if ($key eq $entry{title}) {
			# group name is not defined. read it from the .MAN file.
			while ($_ = $entry{handle}->getline()) {
				s/[ \r\n]+$//;
				if (s/^ +([^ ]+p̊bm) +[0-9]+$/$1/) {
					$entry{title} = &reform_sjis($_, 0);
					$entry{title} =~ s/ //g;
					$wdic->{wdic_group}{$key}{NAME} = $entry{title};
					last;
				}
			}
		}

		$refs{$key} = { %entry };
		&write_link($fpwmenu, "$entry{title}\n", $entry{tag});
	}

	foreach $key (@keys) {
		%entry = %{ $refs{$key} };
		&log("  $entry{file}\n");
		&new_menu_context("$LABEL_SJIS{ref} ($entry{title})", $entry{tag});
		while ($_ = $entry{handle}->getline()) {
			if (/^\-+  (b)  \-+$/) {
				&write_man_section_references($entry{handle});
				last;
			}
		}
		$entry{handle}->close();
	}
}

#
# void write_question(handle)
# creates a new menu entry from the QUESTION.TXT file. not tested any longer.
#
sub write_question($)
{
	my $handle = $_[0];
	local $_;

	&log("$wdic_dir/QUESTION.TXT\n");
	$handle->seek(0, IO::File::SEEK_SET);

	while ($_ = $handle->getline()) {
		last if ($_ =~ s/^\s+([^\r\n]+)[\r\n]+$/$1/);
	}

	&new_menu_context($_, 'QUESTION');
	while ($_ = $handle->getline()) {
		if (/^\- \- \-/) {
			&write_sjis($fpwmenu, $_);
			last;
		}
		s/^ {5}//;
		s/([^\r\n]*)[\r\n]+$/$1/ if (!/^ / && !/B\r\n$/);
		&write_sjis($fpwmenu, $_);
	}
	while ($_ = $handle->getline()) {
		&write_sjis($fpwmenu, $_);
		last if (/^\- \- \-/);
	}
	while ($_ = $handle->getline()) {
		s/^ {5}//;
		&write_sjis($fpwmenu, $_)
	}
}

##------------------------------------------------------------------------
## INDEX

#
# void make_group_index(char **file)
# reads *.TAG files and converts them into index entries.
#
sub make_group_index(@)
{
	my (@list, %group, $i, $key, $cat);

	&log("TAG files found, creating group index...\n");

	@list = $wdic->get_group_list();
	for ($i = 0; $i < scalar(@list); $i++) {
		$group{$list[$i]} = $i;
	}
	@list = sort {
		my $a1 = uc($a);
		my $b1 = uc($b);
		$a1 =~ s/^.*\/([^\/]+)\.TAG/$1/;
		$b1 =~ s/^.*\/([^\/]+)\.TAG/$1/;
		if (defined($group{$a1})) {
			defined($group{$b1}) ? $group{$a1} <=> $group{$b1} : 1;
		} elsif (defined($group{b1})) {
			-1;
		} else {
			$a1 cmp $b1;
		}
	} @_;

	undef %group;
	&new_menu_context('', 'GROUP');

	# write the links to each group.
	foreach (@list) {
		my %hash = (
			HANDLE => new IO::File($_, 'r'),
			TAG => '',
			NAME => ''
		);

		$hash{HANDLE} || die "Unable to open file: $_ ($!)";
		binmode $hash{HANDLE};

		$_ = uc($_);
		s/^.*\/([^\/]+)\.TAG/$1/;

		$hash{TAG} = $_ . "_TAG";
		$hash{NAME} = $wdic->get_group_title($_);
		$group{$_} = \%hash;

		&write_link($fpwmenu, "$hash{NAME}\n", $hash{TAG});
	}

	# list up the words in the .TAG file.
	foreach $key (keys(%group)) {
		my %hash = %{ $group{$key} };
		print "$key.TAG: $hash{NAME}\n ";

		# create heading index.
		&new_menu_context('', $hash{TAG});
		&write_link($fpwmenu, "($LABEL_SJIS{index})\n", 'GROUP');

		while ($_ = $hash{HANDLE}->getline()) {
			last if (/^\-+[\r\n]+$/);
			next if (!s/^[A-Z]+\.TAG\([0-9]+\): +y (.+) z.*$/$1/);
			if ('`' eq $_) {
				$_ = 'At@xbg';
			} elsif (/^\x82[\xa0\xa9\xb3\xbd\xc8\xcd\xdc\xe2\xe7]$/) {
				# [Ȃ͂܂]
				$_ .= 's';
			} elsif ('' eq $_) {
				$_ .= 's, ';
			} elsif (length($_) < 4) {
				next;
			}
			&write_link($fpwmenu, "$_\n",
					sprintf("%s%04X", $hash{TAG}, unpack('S', $_)));
		}

		# create word index.
		while ($_ = $hash{HANDLE}->getline()) {
			s/([^\r\n]*)[\r\n]+$/$1/;
			next if (/^\-*$/);

			if (s/^\-+  (.+)  \-+$/$1/) {
				# heading line.
				print $_;
				$newline_on_death = 1;
				if (/^\x82[\x60\xa0\xa9\xb3\xbd\xc8\xcd\xdc\xe2\xe7\xed]$/
						|| 4 <= length($_)) {
					# [`Ȃ͂܂], L, 
					$_ = 'A' if (/^`$/);
					&new_menu_context('',
							sprintf("%s%04X", $hash{TAG}, unpack('S', $_)));
					&write_link($fpwmenu, "($hash{NAME})\n", $hash{TAG});
				}
				next;
			} elsif (!s/^([A-Z]+)\.DIC\([0-9]+\): +\#(.+)$/$2/) {
				next;
			}
			undef @list;
			$list[0] = $1;		# dictionary.

			s/^(.+[^ ]) +(\[.+)$/$2/ || die "$key.TAG: Unexpected format: $_";
			$list[1] = $1;		# word.

			if (s/(\]) (.+)$/$1/) {
				# reference to another word.
				$list[2] = $2;	# target.
				&write_sjis($fpwmenu, "$list[1] $_");
				&write_word_link($fpwmenu, " ", $list[2], $key, '');
			} else {
				&write_word_link($fpwmenu, '', $list[1], $key, '');
				&write_sjis($fpwmenu, " $_");
			}

			$cat = $wdic->get_file_title($list[0]);
			$cat = &reform_sjis($cat, 0);
			$cat =~ s/ //g;
			&write_sjis($fpwmenu, " s$catt\n");
		}

		if ($newline_on_death) {
			print "\n";
			$newline_on_death = 0;
		}
	}
}

#
# void make_dir_index(%dir)
# creates a menu entry for a specified directory. the first argument should
# be a string that uniquely identifies the directory.
#
sub make_dir_index(;%)
{
	my (%dir, @child);
	local $_;

	if (@_) {
		%dir = @_;
		print "  $dir{DIR}\n";
		&new_menu_context('', $dir{TAG});
		&write_dir_link($fpwmenu, 1, $dir{DIR});
	} else {
		print Jcode::convert($LABEL_SJIS{dir}, $Wdic::CHARSET), "\n";
		&new_menu_context('', 'DIR');
	}

	# links to the child directories.
	@child = $wdic->get_dir_list(%dir ? $dir{DIR} : '/');

	for (my $i = 0; $i <= $#child; $i++) {
		my %info = $wdic->get_dir_info($child[$i]);
		%info || die "Unable to get directory info: $child[$i]";
		if (!$fpwmenu->add_text($LABEL_EUC{bull_dir})
				|| !$fpwmenu->add_reference_start()
				|| !$fpwmenu->add_text($info{NAME_EUC})
				|| !$fpwmenu->add_reference_end($info{TAG})
				|| !$fpwmenu->add_newline()) {
			die &fpwerr("Unable to write directory reference: " .
					"$LABEL_EUC{bull_dir}$info{NAME_EUC} ($info{TAG})",
					$fpwmenu);
		}
		if ($info{REF}) {
			undef $child[$i];
		} else {
			$child[$i] = \%info;
		}
	}

	# links to the member words.
	if (!%dir) {
		# nop.
	} elsif ($dir{LIST}) {
		foreach (@{ $dir{LIST} }) {
			$fpwmenu->add_text($LABEL_EUC{bull_word})
				|| die &fpwerr("Unable to write text: $LABEL_EUC{bull_word}",
						$fpwmenu);
			&write_word_link($fpwmenu, '', "$_\n");
		}
	} elsif (!@child
			&& (!$fpwmenu->add_text($LABEL_EUC{none})
			|| !$fpwmenu->add_newline())) {
		die &fpwerr("Unable to write text: $LABEL_EUC{none}\\n", $fpwmenu);
	}

	# create new entries for child diretcories.
	foreach (@child) {
		&make_dir_index(%{ $_ }) if ($_);
	}
}

##------------------------------------------------------------------------
## PLUGIN LIST

#
# void make_plugin_list(void)
# creates a list of plugins embedded into the body. this function must be
# called after %cgraphs and %plugins have been filled.
#
sub make_plugin_list()
{
	my (@files, $key);

	if ($opt_cgraphs) {
		my (%graphs, $index);

		print "Creating graphics list...";
		$newline_on_death = 1;

		foreach $key (keys(%cgraphs)) {
			if ($cgraphs{$key}{name}) {
				$graphs{$cgraphs{$key}{name}} = "GRAPH_$key";
				push(@files, $cgraphs{$key}{name});
			}
		}

		@files = sort @files;
		&new_menu_context('', 'GRAPH');
		foreach $key (@files) {
			&write_link($fpwmenu, "$key\n", $graphs{$key});
		}

		print " done.\n";
		$newline_on_death = 0;
	}

	if ($opt_plugin) {
		print "Creating plugin list...";
		$newline_on_death = 1;

		@files = sort keys(%plugins);
		&new_menu_context('', 'PLUGIN');
		foreach $key (@files) {
			&write_link($fpwmenu, "$key\n", $plugins{$key}{tag});
		}

		print " done.\n";
		$newline_on_death = 0;
	}
}

##------------------------------------------------------------------------
## COPYRIGHT

#
# void make_copyright(void)
# creates a copyright entry using WDIC.MAN information and conversion
# date/time.
#
sub make_copyright()
{
	local $_;

	foreach (@copyright) {
		&write_sjis($fpwcopyright, "$_\n") if ($_);
	}
	&write_newline($fpwcopyright, 1);
	$_ = sprintf("This book was generated by %s on %s using wdic-fpw %s.\n",
			&user(), &now(), $VERSION =~ /([0-9\.]+)/ && $1);
	$fpwcopyright->add_text($_);
}

1;
